@@ -10,34 +10,33 @@
 #' As well as acting as a container for multiple filter conditions, the
 #' `PivotFilters` class also contains logic for combining filter.
 #' The `action` parameter in many of the methods controls how two filters
-#' are combined.
-#  Most common cases:
-#  (1) When working out the rowColFilters for each pivot table cell, the
-#      filters from the row and column leaf groups are combined using
-#      `action="intersect"`.
-#  (2) When combining the rowColFilters with calculation filters the
-#      action could be any of (in order of most typical)
-#      "intersect", "replace" or "union".
-#      "intersect" would apply additional restrictions, e.g. see the
-#      example in the Calculations vignette that has a measure for
-#      weekend trains only.
-#      "replace" would apply when doing things like percentage of row
-#      total calculations - again, see example in the calculations vignette
-#      "union" is probably much less likely (hard to envisage many
-#      situations when that would be needed).
-#  (3) In custom calculation functions, the action could be any of
-#      "intersect", "replace" or "union".
-#  NOTE: `pivottabler` does not allow complex conditions to be built up,
-#      such as ((A=X) or (B=Y)) and (C=2) since there is complex precedence
-#      involved and conditions like this are not typical of pivot tables.
-#      If they were really needed, a workaround would be to use a custom
-#      calculation function and include this logic in that function.
-#  See Appendix 2 vignette for many more complex calculation details.
+#' are combined.\cr
+#' Most common cases:\cr
+#' (1) When working out the rowColFilters for each pivot table cell, the
+#'     filters from the row and column leaf groups are combined using
+#'     `action="intersect"`.\cr
+#' (2) When combining the rowColFilters with calculation filters the
+#'     action could be any of (in order of most typical)
+#'     "intersect", "replace" or "union".\cr
+#'     "intersect" would apply additional restrictions, e.g. see the
+#'     example in the Calculations vignette that has a measure for
+#'     weekend trains only.\cr
+#'     "replace" would apply when doing things like percentage of row
+#'     total calculations - again, see example in the calculations vignette\cr
+#'     "union" is probably much less likely (hard to envisage many
+#'     situations when that would be needed).\cr
+#' (3) In custom calculation functions, the action could be any of
+#'     "intersect", "replace" or "union".\cr
+#' NOTE: `pivottabler` does not allow complex conditions to be built up,
+#'     such as ((A=X) or (B=Y)) and (C=2) since there is complex precedence
+#'     involved and conditions like this are not typical of pivot tables.
+#'     If they were really needed, a workaround would be to use a custom
+#'     calculation function and include this logic in that function.\cr
+#' See Appendix 2 vignette for many more complex calculation details.
 #'
 #' @docType class
 #' @importFrom R6 R6Class
 #' @importFrom data.table data.table is.data.table
-#' @import jsonlite
 #' @export
 #' @format \code{\link{R6Class}} object.
 #' @examples
@@ -163,15 +162,15 @@ PivotFilters <- R6::R6Class("PivotFilters",
 
     #' @description
     #' Tests whether this `PivotFilters` object matches specified criteria.
-    #' @param matchMode Either "simple" (default) or "combinations".
+    #' @param matchMode Either "simple" (default) or "combinations".\cr
     #' "simple" is used when matching only one variable-value, multiple
     #' variable-value combinations are effectively logical "OR", i.e.
     #' any one single `PivotFilter` match means the `PivotFilters` object
-    #' is a match.
+    #' is a match.\cr
     #' "combinations" is used when matching for combinations of variable
     #' values, multiple variable-value combinations are effectively
     #' logical "AND", i.e. there must be a matching `PivotFilter` for
-    #' every variable name / variable values criteria specified.
+    #' every variable name / variable values criteria specified.\cr
     #' See the "Finding and Formatting" vignette for graphical examples.
     #' @param variableNames The variable name(s) to find a filter for.  This
     #' can be a vector containing more than one variable name.
@@ -515,7 +514,16 @@ PivotFilters <- R6::R6Class("PivotFilters",
     #' @description
     #' Return the contents of this object as JSON for debugging.
     #' @return A JSON representation of various object properties.
-    asJSON = function() { return(jsonlite::toJSON(self$asList())) },
+    asJSON = function() {
+      if (!requireNamespace("jsonlite", quietly = TRUE)) {
+        stop("The jsonlite package is needed to convert to JSON.  Please install the jsonlite package.", call. = FALSE)
+      }
+      jsonliteversion <- utils::packageDescription("jsonlite")$Version
+      if(numeric_version(jsonliteversion) < numeric_version("1.1")) {
+        stop("Version 1.1 or above of the jsonlite package is needed to convert to JSON.  Please install an updated version of the jsonlite package.", call. = FALSE)
+      }
+      return(jsonlite::toJSON(self$asList()))
+    },
 
     #' @description
     #' Return a representation of this object as a character value.
